import os
import sys
import urllib.parse

from testinfra import backend, host, modules
from testinfra.backend import ssh, docker
from testinfra.utils import cached_property
from testinfra.modules.base import InstanceModule

from remork import client, process, files, utils


class attrdict(dict):
    __getattr__ = dict.__getitem__


class RemorkBackendMixin(object):
    def __init__(self, *args, **kwargs):
        self.remork_config = {
            'python': kwargs.pop('remork_python', 'python'),
        }
        super().__init__(*args, **kwargs)

    def run(self, command, *args, **kwargs):
        command = self.get_command(command, *args)
        rv = process.run_helper(self.router, command, **kwargs)
        exit_code, stdout, stderr = rv.wait()
        result = self.result(exit_code, command, stdout, stderr)
        return result


class SshBackend(RemorkBackendMixin, ssh.SshBackend):
    NAME = "remork+ssh"

    @cached_property
    def router(self):
        cmd, args = self._build_ssh_command(self.remork_config['python'])
        command = self.quote(' '.join(cmd), *args)
        return client.connect(command, double_quote=True)


class DockerBackend(RemorkBackendMixin, docker.DockerBackend):
    NAME = "remork+docker"

    @cached_property
    def router(self):
        if self.user:
            shell = self.get_command(
                'docker exec -i -u %s %s /bin/sh -c {cmd}', self.user, self.name)
        else:
            shell = self.get_command('docker exec -i %s /bin/sh -c {cmd}', self.name)
        return client.connect(self.remork_config['python'], shell_cmd=shell)


class RemorkModule(InstanceModule):
    def upload(self, source=None, dest=None, content=None):
        assert dest
        assert source or content
        rv = None
        if source:
            if os.path.isfile(source):
                if dest[-1] == os.sep:
                    dest = dest + os.path.basename(source)
                if hasattr(source, 'read'):
                    rv = files.upload_file_helper(self._host.backend.router, dest, source=source)
                else:
                    with open(source, 'rb') as fd:
                        rv = files.upload_file_helper(self._host.backend.router, dest, source=fd)
            elif os.path.isdir(source):
                root = os.path.basename(source.rstrip(os.sep))
                fsource = source
                if root == '.':
                    root = ''
                else:
                    fsource = os.path.dirname(source)
                for fname in utils.walkdir(source, root):
                    sfname = os.path.join(fsource, fname)
                    with open(sfname, 'rb') as fd:
                        rv = files.upload_file_helper(
                            self._host.backend.router,
                            os.path.join(dest, fname),
                            source=fd, mode=os.fstat(fd.fileno()).st_mode)
            else:
                raise OSError('only file and dir sources are supported')
        else:
            rv = files.upload_file_helper(self._host.backend.router, dest, content=content)

        if rv:
            rv.wait()

    def lineinfile(self, path, line):
        rv = self._host.backend.router.call('remork.files', 'lineinfile', path, line)
        return attrdict({'changed': rv.wait()})

    def blockinfile(self, path, marker, block):
        rv = self._host.backend.router.call('remork.files', 'blockinfile', path, marker, block)
        return attrdict({'changed': rv.wait()})


modules.modules['remork'] = 'remork:RemorkModule'
sys.modules['testinfra.modules.remork'] = sys.modules['remork.testinfra']


class HostMixin:
    def run_check(self, *args, **kwargs):
        return self.run_expect([0], *args, **kwargs)


host.Host.run_check = HostMixin.run_check

backend.BACKENDS[SshBackend.NAME] = 'remork.testinfra.SshBackend'
backend.BACKENDS[DockerBackend.NAME] = 'remork.testinfra.DockerBackend'
BACKENDS = ['ssh', 'docker']


def init(spec):
    backends = set(filter(None, (it.strip() for it in spec.split())))
    for it in backends:
        if it in BACKENDS:
            backend.BACKENDS[it] = backend.BACKENDS[f'remork+{it}']


def command_result_repr(self):
    def decode(data):
        if type(data) is type(''):
            return data
        return data.decode('utf-8', errors='ignore')

    def indent(text, prefix):
        return '\n'.join(prefix + it for it in text.splitlines())

    out = []
    if self._stdout_bytes:
        out.append('  STDOUT:\n' + indent(decode(self._stdout_bytes.rstrip()), '    '))
    if self._stderr_bytes:
        out.append('  STDERR:\n' + indent(decode(self._stderr_bytes.rstrip()), '    '))
    if out:
        out = [','] + out + ['']

    return (
        "CommandResult(command=%s, exit_status=%s%s)"
    ) % (
        repr(self.command),
        self.exit_status,
        '\n'.join(out),
    )

backend.base.CommandResult.__repr__ = command_result_repr


parse_host_spec_orig = backend.parse_hostspec
def parse_hostspec(hostspec):
    host, kw = parse_host_spec_orig(hostspec)

    if hostspec is not None and "://" in hostspec:
        url = urllib.parse.urlparse(hostspec)
        query = urllib.parse.parse_qs(url.query)
        for k, v in query.items():
            if k not in kw:
                kw[k] = v[0]
    return host, kw

backend.parse_hostspec = parse_hostspec


if os.environ.get('REMORK_OVERRIDE_BACKEND'):
    init(os.environ['REMORK_OVERRIDE_BACKEND'])



import subprocess
from remork.router import bg, drain

procs = {}


def out_drain_handler(router, msg_id, data_type):
    def handler(data):
        if data:
            router.write_data(msg_id, data_type, data)
    return handler


def wait_proc(router, msg_id, proc, out_t, err_t):
    out_t.join()
    err_t.join()
    proc.wait()
    procs.pop(msg_id, None)
    router.done(msg_id, proc.returncode)


def run(router, msg_id, command, shell=True, cwd=None, env=None, has_stdin=False):
    p = subprocess.Popen(
        command, shell=shell, cwd=cwd, env=env,
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=0)
    procs[msg_id] = p
    out_t = bg(drain, p.stdout, out_drain_handler(router, msg_id, 1))
    err_t = bg(drain, p.stderr, out_drain_handler(router, msg_id, 2))
    bg(wait_proc, router, msg_id, p, out_t, err_t)

    if has_stdin:  # pragma: no cover
        def handler(data_type, data):
            if data_type:
                if msg_id in procs:
                    procs[msg_id].stdin.write(data)
            else:
                if msg_id in procs:
                    procs[msg_id].stdin.close()

        router.data_subscribe(msg_id, handler)


#==LOCAL==
def on_result(result, value):
    return (value, b''.join(result.data.get(1, [])), b''.join(result.data.get(2, [])))


def run_helper(router, command, env=None, cwd=None, shell=True):
    return router.call('remork.process', 'run', command, env=env, cwd=cwd, shell=shell, on_result_=on_result)



import os.path
from remork.router import bstr, nstr, btype, debug, simplecall


def upload_file(router, msg_id, dest, mode=None):
    dest = nstr(dest)
    if not os.path.exists(os.path.dirname(dest)):
        os.makedirs(os.path.dirname(dest))

    f = open(dest, 'wb')
    if mode:
        os.fchmod(f.fileno(), mode)

    def handler(data_type, data):
        if data_type:
            f.write(data)
        else:
            f.close()
            router.done(msg_id)

    router.data_subscribe(msg_id, handler)


def read_file(path, default=None):
    if os.path.exists(path):
        with open(path) as fd:
            return fd.read()
    return default


def atomic_write(path, content):
    if type(content) is btype:
        mode = 'wb'
    else:
        mode = 'w'
    tmp = path + '.remork-tmp'
    with open(tmp, mode) as fd:
        fd.write(content)
    os.rename(tmp, path)


@simplecall
def lineinfile(path, line):
    line = nstr(line)
    path = nstr(path)
    lines = read_file(path, '').splitlines()
    found = False
    for it in lines:
        if it == line:
            found = True
            break

    if not found:
        lines.append(line+'\n')
        atomic_write(path, '\n'.join(lines))

    return not found


def find_line(lines, line):
    try:
        return lines.index(line)
    except ValueError:
        return None


@simplecall
def blockinfile(path, marker, block):
    path = nstr(path)
    marker = nstr(marker)
    block = nstr(block)
    startmarker = marker + ' REMORK BLOCK START'
    endmarker = marker + ' REMORK BLOCK END'

    content = read_file(path, '')
    lines = content.splitlines()

    head = lines
    tail = []
    start = find_line(lines, startmarker)
    if start is not None:
        end = find_line(lines, endmarker)
        if end is not None:
            head = lines[:start]
            tail = lines[end+1:]

    lines = head + [startmarker, block.rstrip('\n'), endmarker] + tail + ['']
    newcontent = '\n'.join(lines)
    changed = content != newcontent
    if changed:
        atomic_write(path, newcontent)
    return changed


#==LOCAL==
from remork.router import iter_read

def upload_file_helper(router, dest, source=None, content=None, mode=None):
    rv = router.call('remork.files', 'upload_file', dest=dest, mode=mode)
    if content is not None:
        rv.write_data(1, bstr(content))
    elif source is not None:
        for data in iter_read(source, 1 << 18):
            rv.write_data(1, data, compress=len(data) > 512)
    rv.end_data()
    return rv



from __future__ import print_function

import os
import sys
import time
import struct
import json
import zlib
import threading
import itertools
import traceback

# PY2 = sys.version_info.major <= 2
# if PY2:
#     from StringIO import StringIO as BytesIO
# else:
#     from io import BytesIO

DEBUG = int(os.environ.get('REMORK_DEBUG', '0'))
DEBUG_LOG = None # open('/tmp/boo.log', 'w')
START = None

msg_hdr_st = struct.Struct('!Q B I')  # descriptor message-type message-subtype payload-size
IOSIZE = 65536
MSG_BUFFER_SIZE = 1 << 20
SIDE = 'REMOTE'

btype = type(b'')
ntype = type('')


def bstr(s, encoding='utf-8'):
    if type(s) != btype:
        s = s.encode(encoding)
    return s


def nstr(s, encoding='utf-8'):
    stype = type(s)
    if stype != ntype:
        if stype == btype:
            s = s.decode(encoding)
        else:  # pragma: no cover
            s = s.encode(encoding)
    return s


CALL = 1
RESULT = 2
DATA = 3
DATA_FLAG = 0x40
COMPRESS_FLAG = 0x80


class StopDrain(Exception):
    def __init__(self, value):  # pragma: no cover
        self.value = value


def bg(fn, *args, **kwargs):
    log = kwargs.pop('log_', None)
    def wrapper():
        try:
            fn(*args, **kwargs)
        except:  # pragma: no cover
            if log:
                log.exception('{} router error'.format(SIDE))
            elif traceback is not None:
                traceback.print_exc()

    t = threading.Thread(target=wrapper)
    t.daemon = True
    t.start()
    return t


def copydata(dest, data):
    while data:
        size = dest.write(data)
        debug('writedata', dest, size, data[:size])
        data = data[size:]


class Buffer(object):
    def __init__(self, size):
        lock = threading.RLock()
        self.need_space = threading.Condition(lock)
        self.need_data = threading.Condition(lock)
        self.buffer = b''
        self.maxsize = size
        self.is_done = False

    def done(self):  # pragma: no cover
        with self.need_data:
            self.is_done = True
            self.need_data.notify()

    def fresh(self):
        return Buffer(self.maxsize)

    def write(self, data):
        debug('buf-write', len(self.buffer), data, level=3)
        dsize = len(data)
        assert dsize <= self.maxsize
        with self.need_space:
            while dsize > self.maxsize - len(self.buffer):  # pragma: no cover
                debug('buf-write-wait', len(data), len(self.buffer), level=3)
                self.need_space.wait(0.1)

            self.buffer += data
            self.need_data.notify()
        return dsize

    def read(self, size):
        debug('buf-read-start', level=3)
        with self.need_data:
            debug('buf-read-lock', level=3)
            while not self.buffer and not self.is_done:
                debug('buf-read-wait', len(self.buffer), level=3)
                self.need_data.wait(0.01)

            result = self.buffer[:size]
            debug('buf-read-get', result, level=3)
            self.buffer = self.buffer[size:]
            self.need_space.notify_all()

        return result


class Reader(object):
    def start(self, receiver):
        self.receiver = receiver
        receiver.send(None)
        gen = self.feeder()
        gen.send(None)
        return gen.send

    def read(self, size):
        self.size = size

    def feeder(self):
        buf = b''
        while True:
            data = yield
            buf += data
            while len(buf) >= self.size:
                sz = self.size
                self.receiver.send(buf[:sz])
                buf = buf[sz:]


def protocol(fn):
    def inner(*args, **kwargs):
        reader = Reader()
        receiver = fn(reader, *args, **kwargs)
        return reader.start(receiver)
    return inner


@protocol
def msg_proto_decode(reader, handler):
    while True:
        data = yield reader.read(msg_hdr_st.size)
        msg_id, msg_type, size = msg_hdr_st.unpack_from(data)

        data = yield reader.read(size)
        debug('proto', handler, msg_id, msg_type, data, level=2)

        if msg_type & COMPRESS_FLAG:
            data = zlib.decompress(data)
            msg_type = msg_type & ~COMPRESS_FLAG

        handler(msg_id, msg_type, data)


def msg_encode(msg_id, msg_type, data, compress=False):
    if compress:
        msg_type |= COMPRESS_FLAG
        data = zlib.compress(data)
    return msg_hdr_st.pack(msg_id, msg_type, len(data)) + data


def iter_read(src, size=IOSIZE):
    while True:
        data = src.read(size)
        if not data:
            break
        yield data


class FD(object):  # pragma: no cover
    def __init__(self, fd):
        self.fd = fd

    def read(self, size):
        return os.read(self.fd, size)

    def write(self, data):
        return os.write(self.fd, data)

    def flush(self):
        pass

    def __repr__(self):
        return 'FD({0})'.format(self.fd)


def iter_write(dest, iterator, size=IOSIZE, close=False, flush=False):
    buf = b''
    for data in iterator:
        debug('write, getdata', dest, data, level=3)
        buf += data
        is_first = True
        while is_first or len(buf) > size:
            written = dest.write(buf[:size])
            debug('write', dest, written, buf[:size], level=4)
            buf = buf[written:]
            is_first = False

    if flush:  # pragma: no cover
        dest.flush()

    if close:  # pragma: no cover
        dest.close()


def copy(dest, src, size=IOSIZE, close=False):
    it = iter_read(src, size)
    iter_write(dest, it, size, close=close)


def drain(src, handler, size=IOSIZE):
    for data in iter_read(src, size):
        debug('read', src, data, level=3)
        try:
            handler(data)
        except StopDrain as e:  # pragma: no cover
            return e.value

    handler(b'')


def debug(*args, **kwargs):
    force = kwargs.get('force', False)
    level = kwargs.get('level', 2)
    if DEBUG >= level or force or DEBUG_LOG:  # pragma: no cover
        args = list(args)
        now = time.time()

        if START:  # pragma: no cover
            targs = [now, round((now - START)*1000)]
        else:
            targs = [now]

        if not force and type(args[-1]) == type(b'') and len(args[-1]) > 100:
            args[-1] = args[-1][:100] + b'...(%d bytes total)' % len(args[-1])
        if DEBUG_LOG:  # pragma: no cover
            print(SIDE + ':', *(targs + args), file=DEBUG_LOG)
            DEBUG_LOG.flush()
        else:
            print(SIDE + ':', *(targs + args), file=sys.stderr)
            sys.stderr.flush()


def simplecall(fn):
    fn._remork_simple = True
    return fn


class Router(object):
    def __init__(self, bufsize=None):
        self.buffer = Buffer(bufsize or MSG_BUFFER_SIZE)
        self.write = self.buffer.write
        self.results = {}
        self.counter = itertools.count(1)
        self.result_cond = threading.Condition()

    def reset(self):
        self.buffer = self.buffer.fresh()
        self.write = self.buffer.write

    def make_result(self, on_result):
        msg_id = next(self.counter)
        result = self.results[msg_id] = Result(self, msg_id, on_result)
        return result

    def write_msg(self, msg_id, msg_type, data, compress=None):
        self.buffer.write(msg_encode(msg_id, msg_type, data, compress=compress))

    def done(self, msg_id, result=None):
        self.results.pop(msg_id, None)
        self.write_msg(msg_id, RESULT, bstr(json.dumps({'result': result})))

    def error(self, msg_id, error):
        self.results.pop(msg_id, None)
        self.write_msg(msg_id, RESULT, bstr(json.dumps({'error': error})))

    def data_subscribe(self, msg_id, handler):
        self.results[msg_id] = handler

    def write_data(self, msg_id, data_type, data, compress=None):
        self.write_msg(msg_id, DATA_FLAG | data_type, data, compress=compress)

    def end_data(self, msg_id):
        self.write_msg(msg_id, DATA_FLAG, b'')

    def forget(self, msg_id):  # pragma: no cover
        self.results.pop(msg_id, None)

    def handle_msg(self, msg_id, msg_type, data):
        if msg_type & DATA_FLAG:
            data_type = msg_type & ~DATA_FLAG
            msg_type = DATA
        try:
            if msg_type == CALL:
                info = json.loads(nstr(data))
                m = sys.modules[info['module']]
                fn = getattr(m, info['fn'])
                if getattr(fn, '_remork_simple', None):
                    result = fn(*info['args'], **info.get('kwargs', {}))
                    self.done(msg_id, result)
                else:
                    fn(self, msg_id, *info['args'], **info.get('kwargs', {}))
            elif msg_type == RESULT:
                debug('results', self.results)
                info = json.loads(nstr(data))
                r = self.results.pop(msg_id, None)
                if r:
                    if 'error' in info:
                        r.set_error(info['error'])
                    else:
                        r.set_result(info['result'])
                with self.result_cond:
                    self.result_cond.notify_all()
            elif msg_type == DATA:
                r = self.results.get(msg_id)
                if r:
                    if hasattr(r, 'feed'):
                        r.feed(data_type, data)
                    else:
                        r(data_type, data)
        except Exception as e:
            import traceback
            traceback.print_exc()
            self.error(msg_id, str(e))


def router():  # pragma: no cover
    debug('starting router', level=2)
    r = Router()
    r.buffer.write(FD(0).read(10))
    bg(copy, FD(1), r.buffer)
    proto = msg_proto_decode(r.handle_msg)
    debug('router started', level=1)
    drain(FD(0), proto)


def inject_modules(router, msg_id, modules):  # pragma: no cover
    for minfo in modules:
        debug('injecting', minfo['module'], id(router), level=1)
        mname = nstr(minfo['module'])
        pparts = mname.split('.')[:-1]
        mod = type(sys)(mname)
        mod.__file__ = nstr(minfo['file'])
        if pparts:
            mod.__package__ = '.'.join(pparts)
            p = []
            for part in pparts[:-1]:
                p.append(part)
                tp = '.'.join(p)
                if tp not in sys.modules:
                    sys.modules[tp] = type(sys)(tp)

        code = compile(nstr(minfo['source']), mod.__file__, 'exec')
        exec(code, vars(mod))
        sys.modules[mname] = mod


if __name__ == '__main__':  # pragma: no cover
    import sys
    sys.modules['remork'] = type(sys)('remork')
    sys.modules['remork.router'] = sys.modules['__main__']


#==LOCAL==
SIDE = 'LOCAL'

class ResultException(Exception):
    pass


class Result:
    def __init__(self, router, msg_id, on_result=None):
        self.router = router
        self.msg_id = msg_id
        self.result = None
        self.error = None
        self.data = {}
        self.on_result = on_result
        self.done = False

    def ignore(self):  # pragma: no cover
        self.router.forget(self.msg_id)

    def wait(self):
        while not self.done:
            with self.router.result_cond:
                while not self.done:
                    self.router.result_cond.wait()
        if self.error:
            raise ResultException(self.error)
        return self.result

    def set_result(self, value):
        if self.on_result:
            self.result = self.on_result(self, value)
        else:
            self.result = value
        self.done = True

    def set_error(self, value):
        self.error = value
        self.done = True

    def write_data(self, data_type, data, compress=False):
        self.router.write_data(self.msg_id, data_type, data, compress)

    def end_data(self):
        self.router.end_data(self.msg_id)

    def feed(self, data_type, data):
        self.data.setdefault(data_type, []).append(data)


def msg_call(msg_id, module, fname, args, kwargs={}, compress=False):
    data = bstr(json.dumps({'module': module, 'fn': fname, 'args': args, 'kwargs': kwargs}))
    return msg_encode(msg_id, CALL, data, compress=compress)


class LocalRouter(Router):
    def check(self):
        pass

    def call(self, module, fname, *args, **kwargs):
        self.check()
        compress = kwargs.pop('compress_', False)
        on_result = kwargs.pop('on_result_', None)
        result = self.make_result(on_result)
        self.buffer.write(msg_call(result.msg_id, module, fname, args, kwargs, compress=compress))
        return result



from setuptools import setup, find_packages
import remork

setup(
    name='remork',
    version=remork.VERSION,
    url='https://github.com/baverman/remork/',
    license='MIT',
    author='Anton Bobrov',
    author_email='baverman@gmail.com',
    description='Configuration management and accelerator for testinfra',
    long_description=open('README.md').read(),
    packages=find_packages(exclude=['tests']),
    zip_safe=False,
    platforms='any',
    install_requires=[
        'testinfra',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Systems Administration',
    ]
)



import os.path


def walkdir(top, root='', follow_symlinks=False):
    with os.scandir(top) as iterator:
        dirs = []
        for entry in iterator:
            if entry.is_dir():
                if not follow_symlinks and entry.is_symlink():
                    continue
                dirs.append(entry.name)
            else:
                yield os.path.join(root, entry.name)

        for d in dirs:
            yield from walkdir(os.path.join(top, d), os.path.join(root, d))



import os.path
import zlib
import shlex
import subprocess
import time
import logging

from . import process, files, router
from .router import copy, copydata, msg_proto_decode, drain, bg, bstr, LocalRouter, msg_call

log = logging.getLogger('remork.client')

BOOT_TAIL = '''\
DEBUG = %d
router()
'''

MODULES = [process, files]


class ConnectException(OSError): pass


def extract_remote_part(source):
    return source.partition('#==LO' + 'CAL==')[0]


def make_source_bundle(modules):
    result = []
    for m in modules:
        fname = m.__file__
        if fname.endswith('.pyc'):  # pragma: no cover
            fname = fname[:-4] + '.py'
        with open(fname, 'r') as fobj:
            source = extract_remote_part(fobj.read())
        result.append({'module': m.__name__, 'file': m.__file__, 'source': source})
    return result


STAGE0 = '''\
import os, zlib
data = os.read(0, {0})
source = zlib.decompress(data).decode('utf-8')
exec(source)
'''


def bootstrap(debug=0):
    script = []
    fname = os.path.join(os.path.dirname(__file__), 'router.py')
    script.append(extract_remote_part(open(fname).read()))
    script.append(BOOT_TAIL % debug)

    stage1 = zlib.compress(bstr('\n'.join(script)))
    # BAD. here we hope that os.read returns all requested data in a single call
    stage0 = 'import os, zlib; exec(zlib.decompress(os.read(0, {0})).decode("utf-8"))'.format(len(stage1))
    # stage0 = STAGE0.format(len(stage1))
    return stage0, stage1


class ProcessRouter(LocalRouter):
    def __init__(self, cmd, stage1, sources, shell=False, bufsize=None):
        super().__init__(bufsize=bufsize)
        self.stage1 = stage1
        self.sources = sources
        self.cmd = cmd
        self.shell = shell

    def connect(self):
        self.reset()
        self.proc = subprocess.Popen(self.cmd, shell=self.shell, stdin=subprocess.PIPE,
                                     stdout=subprocess.PIPE, stderr=subprocess.PIPE, bufsize=0)

    def handshake(self):
        ack = os.urandom(10)
        call = msg_call(0, 'remork.router', 'inject_modules', (self.sources,), compress=True)
        payload = b''.join([self.stage1, ack, call])
        copydata(self.proc.stdin, payload)
        # TODO: add timeout
        remote_ack = self.proc.stdout.read(10)
        if ack != remote_ack:
            raise ConnectException(
                'Invalid ack: \n' + repr(self.proc.stdout.read()) + '\n' + repr(self.proc.stderr.read()))
        router.debug('handshake done', level=1)

    def check(self):
        if self.proc is None:
            self.connect()
            self.process()

    def monitor(self):
        while True:
            if self.proc.poll() is not None:
                self.close()
                break
            time.sleep(0.9)

    def print_err(self, data):
        if data:
            log.error('REMOTE stderr: ' + repr(data))

    def process(self):
        self.handshake()
        proto = msg_proto_decode(self.handle_msg)
        bg(drain, self.proc.stdout, proto, log_=log)
        bg(drain, self.proc.stderr, self.print_err, log_=log)
        bg(copy, self.proc.stdin, self.buffer, log_=log)
        bg(self.monitor, log_=log)

    def close(self):
        router.debug('Closing local router', level=1)
        self.buffer.done()
        proc = self.proc
        self.proc = None
        for r in self.results.values():
            r.set_error('Closed: ' + repr(proc.args) + '\n' + repr(proc.stdout.read()))
            r.done = True
        with self.result_cond:
            self.result_cond.notify_all()


def make_command(python_cmd, stage0, double_quote=False):
    quoted = shlex.quote(stage0)
    if double_quote:
        quoted = shlex.quote(quoted)
    return python_cmd + ' -c ' + quoted


def connect(python_cmd=['python'], shell_cmd=None, double_quote=False, debug=None, additional=None):
    if debug is None:
        debug = router.DEBUG
    else:
        router.DEBUG = debug
    stage0, stage1 = bootstrap(debug=debug)
    shell = False
    if isinstance(python_cmd, (list, tuple)):
        python_cmd += ['-c', stage0]
    else:  # pragma: no cover
        shell = True
        python_cmd = make_command(python_cmd, stage0, double_quote)
        if shell_cmd:
            python_cmd = shell_cmd.replace('{cmd}', shlex.quote(python_cmd))

    sources = make_source_bundle(MODULES + (additional or []))
    log.info('Creating router: %s', python_cmd)

    c = ProcessRouter(python_cmd, stage1, sources, shell=shell)
    c.connect()
    c.process()
    return c
